@@ -1,19 +1,25 @@
 from types import ModuleType
-from typing import Any, Dict, List, Optional, Tuple, Union
+from typing import TYPE_CHECKING, Any, Optional, Union
 
 import pandas as pd
 from absl.logging import logging
 
-from snowflake.ml._internal import platform_capabilities, telemetry
+from snowflake.ml._internal import env, platform_capabilities, telemetry
 from snowflake.ml._internal.exceptions import error_codes, exceptions
 from snowflake.ml._internal.human_readable_id import hrid_generator
 from snowflake.ml._internal.utils import sql_identifier
-from snowflake.ml.model import model_signature, type_hints as model_types
+from snowflake.ml.model import model_signature, target_platform, task, type_hints
 from snowflake.ml.model._client.model import model_impl, model_version_impl
 from snowflake.ml.model._client.ops import metadata_ops, model_ops, service_ops
 from snowflake.ml.model._model_composer import model_composer
+from snowflake.ml.model._model_composer.model_manifest import model_manifest_schema
 from snowflake.ml.model._packager.model_meta import model_meta
+from snowflake.ml.registry._manager import model_parameter_reconciler
 from snowflake.snowpark import exceptions as snowpark_exceptions, session
+from snowflake.snowpark._internal import utils as snowpark_utils
+
+if TYPE_CHECKING:
+    from snowflake.ml.experiment._experiment_info import ExperimentInfo
 
 logger = logging.getLogger(__name__)
 
@@ -39,24 +45,27 @@ class ModelManager:
     def log_model(
         self,
         *,
-        model: Union[model_types.SupportedModelType, model_version_impl.ModelVersion],
+        model: Union[type_hints.SupportedModelType, model_version_impl.ModelVersion],
         model_name: str,
+        progress_status: type_hints.ProgressStatus,
         version_name: Optional[str] = None,
         comment: Optional[str] = None,
-        metrics: Optional[Dict[str, Any]] = None,
-        conda_dependencies: Optional[List[str]] = None,
-        pip_requirements: Optional[List[str]] = None,
-        artifact_repository_map: Optional[Dict[str, str]] = None,
-        target_platforms: Optional[List[model_types.SupportedTargetPlatformType]] = None,
+        metrics: Optional[dict[str, Any]] = None,
+        conda_dependencies: Optional[list[str]] = None,
+        pip_requirements: Optional[list[str]] = None,
+        artifact_repository_map: Optional[dict[str, str]] = None,
+        resource_constraint: Optional[dict[str, str]] = None,
+        target_platforms: Optional[list[type_hints.SupportedTargetPlatformType]] = None,
         python_version: Optional[str] = None,
-        signatures: Optional[Dict[str, model_signature.ModelSignature]] = None,
-        sample_input_data: Optional[model_types.SupportedDataType] = None,
-        user_files: Optional[Dict[str, List[str]]] = None,
-        code_paths: Optional[List[str]] = None,
-        ext_modules: Optional[List[ModuleType]] = None,
-        task: model_types.Task = model_types.Task.UNKNOWN,
-        options: Optional[model_types.ModelSaveOption] = None,
-        statement_params: Optional[Dict[str, Any]] = None,
+        signatures: Optional[dict[str, model_signature.ModelSignature]] = None,
+        sample_input_data: Optional[type_hints.SupportedDataType] = None,
+        user_files: Optional[dict[str, list[str]]] = None,
+        code_paths: Optional[list[str]] = None,
+        ext_modules: Optional[list[ModuleType]] = None,
+        task: type_hints.Task = task.Task.UNKNOWN,
+        experiment_info: Optional["ExperimentInfo"] = None,
+        options: Optional[type_hints.ModelSaveOption] = None,
+        statement_params: Optional[dict[str, Any]] = None,
     ) -> model_version_impl.ModelVersion:
 
         database_name_id, schema_name_id, model_name_id = self._parse_fully_qualified_name(model_name)
@@ -129,6 +138,7 @@ class ModelManager:
             conda_dependencies=conda_dependencies,
             pip_requirements=pip_requirements,
             artifact_repository_map=artifact_repository_map,
+            resource_constraint=resource_constraint,
             target_platforms=target_platforms,
             python_version=python_version,
             signatures=signatures,
@@ -137,36 +147,44 @@ class ModelManager:
             code_paths=code_paths,
             ext_modules=ext_modules,
             task=task,
+            experiment_info=experiment_info,
             options=options,
             statement_params=statement_params,
+            progress_status=progress_status,
         )
 
     def _log_model(
         self,
-        model: model_types.SupportedModelType,
+        model: type_hints.SupportedModelType,
         *,
         model_name: str,
         version_name: str,
+        progress_status: type_hints.ProgressStatus,
         comment: Optional[str] = None,
-        metrics: Optional[Dict[str, Any]] = None,
-        conda_dependencies: Optional[List[str]] = None,
-        pip_requirements: Optional[List[str]] = None,
-        artifact_repository_map: Optional[Dict[str, str]] = None,
-        target_platforms: Optional[List[model_types.SupportedTargetPlatformType]] = None,
+        metrics: Optional[dict[str, Any]] = None,
+        conda_dependencies: Optional[list[str]] = None,
+        pip_requirements: Optional[list[str]] = None,
+        artifact_repository_map: Optional[dict[str, str]] = None,
+        resource_constraint: Optional[dict[str, str]] = None,
+        target_platforms: Optional[list[type_hints.SupportedTargetPlatformType]] = None,
         python_version: Optional[str] = None,
-        signatures: Optional[Dict[str, model_signature.ModelSignature]] = None,
-        sample_input_data: Optional[model_types.SupportedDataType] = None,
-        user_files: Optional[Dict[str, List[str]]] = None,
-        code_paths: Optional[List[str]] = None,
-        ext_modules: Optional[List[ModuleType]] = None,
-        task: model_types.Task = model_types.Task.UNKNOWN,
-        options: Optional[model_types.ModelSaveOption] = None,
-        statement_params: Optional[Dict[str, Any]] = None,
+        signatures: Optional[dict[str, model_signature.ModelSignature]] = None,
+        sample_input_data: Optional[type_hints.SupportedDataType] = None,
+        user_files: Optional[dict[str, list[str]]] = None,
+        code_paths: Optional[list[str]] = None,
+        ext_modules: Optional[list[ModuleType]] = None,
+        task: type_hints.Task = task.Task.UNKNOWN,
+        experiment_info: Optional["ExperimentInfo"] = None,
+        options: Optional[type_hints.ModelSaveOption] = None,
+        statement_params: Optional[dict[str, Any]] = None,
     ) -> model_version_impl.ModelVersion:
         database_name_id, schema_name_id, model_name_id = sql_identifier.parse_fully_qualified_name(model_name)
         version_name_id = sql_identifier.SqlIdentifier(version_name)
 
-        use_live_commit = platform_capabilities.PlatformCapabilities.get_instance().is_live_commit_enabled()
+        # TODO(SNOW-2091317): Remove this when the snowpark enables file PUT operation for snowurls
+        use_live_commit = (
+            not snowpark_utils.is_in_stored_procedure()  # type: ignore[no-untyped-call]
+        ) and platform_capabilities.PlatformCapabilities.get_instance().is_live_commit_enabled()
         if use_live_commit:
             logger.info("Using live commit model version")
         else:
@@ -207,25 +225,65 @@ class ModelManager:
         # User specified target platforms are defaulted to None and will not show up in the generated manifest.
         if target_platforms:
             # Convert any string target platforms to TargetPlatform objects
-            platforms = [model_types.TargetPlatform(platform) for platform in target_platforms]
-
-        if artifact_repository_map:
-            for channel, artifact_repository_name in artifact_repository_map.items():
-                db_id, schema_id, repo_id = sql_identifier.parse_fully_qualified_name(artifact_repository_name)
-
-                artifact_repository_map[channel] = sql_identifier.get_fully_qualified_name(
-                    db_id,
-                    schema_id,
-                    repo_id,
-                    self._database_name,
-                    self._schema_name,
+            platforms = [type_hints.TargetPlatform(platform) for platform in target_platforms]
+        else:
+            # Default the target platform to warehouse if not specified and any table function exists
+            if options and (
+                options.get("function_type") == model_manifest_schema.ModelMethodFunctionTypes.TABLE_FUNCTION.value
+                or (
+                    any(
+                        opt.get("function_type") == "TABLE_FUNCTION"
+                        for opt in options.get("method_options", {}).values()
+                    )
                 )
+            ):
+                logger.info(
+                    "Logging a partitioned model with a table function without specifying `target_platforms`. "
+                    'Default to `target_platforms=["WAREHOUSE"]`.'
+                )
+                platforms = [target_platform.TargetPlatform.WAREHOUSE]
+
+            # Default the target platform to SPCS if not specified when running in ML runtime
+            if not platforms and env.IN_ML_RUNTIME:
+                logger.info(
+                    "Logging the model on Container Runtime for ML without specifying `target_platforms`. "
+                    'Default to `target_platforms=["SNOWPARK_CONTAINER_SERVICES"]`.'
+                )
+                platforms = [target_platform.TargetPlatform.SNOWPARK_CONTAINER_SERVICES]
+
+        reconciler = model_parameter_reconciler.ModelParameterReconciler(
+            database_name=self._database_name,
+            schema_name=self._schema_name,
+            conda_dependencies=conda_dependencies,
+            pip_requirements=pip_requirements,
+            target_platforms=target_platforms,
+            artifact_repository_map=artifact_repository_map,
+            options=options,
+        )
+
+        model_params = reconciler.reconcile()
+
+        # Use reconciled parameters
+        artifact_repository_map = model_params.artifact_repository_map
+        save_location = model_params.save_location
 
         logger.info("Start packaging and uploading your model. It might take some time based on the size of the model.")
+        progress_status.update("packaging model...")
+        progress_status.increment()
+
+        if save_location:
+            logger.info(f"Model will be saved to local directory: {save_location}")
 
         mc = model_composer.ModelComposer(
-            self._model_ops._session, stage_path=stage_path, statement_params=statement_params
+            self._model_ops._session,
+            stage_path=stage_path,
+            statement_params=statement_params,
+            save_location=save_location,
         )
+
+        progress_status.update("creating model manifest...")
+        progress_status.increment()
+
         model_metadata: model_meta.ModelMetadata = mc.save(
             name=model_name_id.resolved(),
             model=model,
@@ -234,6 +292,7 @@ class ModelManager:
             conda_dependencies=conda_dependencies,
             pip_requirements=pip_requirements,
             artifact_repository_map=artifact_repository_map,
+            resource_constraint=resource_constraint,
             target_platforms=platforms,
             python_version=python_version,
             user_files=user_files,
@@ -241,7 +300,11 @@ class ModelManager:
             ext_modules=ext_modules,
             options=options,
             task=task,
+            experiment_info=experiment_info,
         )
+
+        progress_status.update("uploading model files...")
+        progress_status.increment()
         statement_params = telemetry.add_statement_params_custom_tags(
             statement_params, model_metadata.telemetry_metadata()
         )
@@ -249,7 +312,8 @@ class ModelManager:
             statement_params, {"model_version_name": version_name_id}
         )
 
-        logger.info("Start creating MODEL object for you in the Snowflake.")
+        progress_status.update("creating model object in Snowflake...")
+        progress_status.increment()
 
         self._model_ops.create_from_stage(
             composed_model=mc,
@@ -276,6 +340,9 @@ class ModelManager:
             version_name=version_name_id,
         )
 
+        progress_status.update("setting model metadata...")
+        progress_status.increment()
+
         if comment:
             mv.comment = comment
 
@@ -289,13 +356,15 @@ class ModelManager:
                 statement_params=statement_params,
             )
 
+        progress_status.update("model logged successfully!")
+
         return mv
 
     def get_model(
         self,
         model_name: str,
         *,
-        statement_params: Optional[Dict[str, Any]] = None,
+        statement_params: Optional[dict[str, Any]] = None,
     ) -> model_impl.Model:
         database_name_id, schema_name_id, model_name_id = self._parse_fully_qualified_name(model_name)
         if self._model_ops.validate_existence(
@@ -323,8 +392,8 @@ class ModelManager:
     def models(
         self,
         *,
-        statement_params: Optional[Dict[str, Any]] = None,
-    ) -> List[model_impl.Model]:
+        statement_params: Optional[dict[str, Any]] = None,
+    ) -> list[model_impl.Model]:
         model_names = self._model_ops.list_models_or_versions(
             database_name=None,
             schema_name=None,
@@ -342,7 +411,7 @@ class ModelManager:
     def show_models(
         self,
         *,
-        statement_params: Optional[Dict[str, Any]] = None,
+        statement_params: Optional[dict[str, Any]] = None,
     ) -> pd.DataFrame:
         rows = self._model_ops.show_models_or_versions(
             database_name=None,
@@ -355,7 +424,7 @@ class ModelManager:
         self,
         model_name: str,
         *,
-        statement_params: Optional[Dict[str, Any]] = None,
+        statement_params: Optional[dict[str, Any]] = None,
     ) -> None:
         database_name_id, schema_name_id, model_name_id = self._parse_fully_qualified_name(model_name)
 
@@ -368,7 +437,7 @@ class ModelManager:
 
     def _parse_fully_qualified_name(
         self, model_name: str
-    ) -> Tuple[
+    ) -> tuple[
         Optional[sql_identifier.SqlIdentifier], Optional[sql_identifier.SqlIdentifier], sql_identifier.SqlIdentifier
     ]:
         try:
